#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>

struct Item {
	struct Item* next;
	struct Item* rep;
	void* val;
	size_t valSize;
};

struct Set {
	struct Item* head;
	struct Item* tail;
	size_t size;
};

typedef struct Set Set;
typedef struct Item Item;

void Item_init(Item* item, void* val, size_t valSize) {
	item->next = NULL;
	item->rep = malloc(sizeof(Item*));
	item->rep = item;
	item->val = val;
	item->valSize = valSize;
}

void Set_createSet(Set* set, void* initialVal, size_t valSize) {
	Item* item = malloc(sizeof(Item));
	Item_init(item, initialVal, valSize);
	set->head = item;
	set->tail = item;
	set->size = 1;
}

void Set_printMembers(Set* set, bool dividers) {
	if (! dividers) {
		Item* currentItem = set->head;
			printf("%d[r:%d]", *(int*)currentItem->val, *(int*)currentItem->rep->val);
		while (currentItem->next != NULL) {
			currentItem = currentItem->next;
			printf("->%d[r:%d]", *(int*)currentItem->val, *(int*)currentItem->rep->val);
		}
		printf("-| (n=%d)\n", set->size);
	} else {
		Item* currentItem = set->head;
			printf(
				"(%d,%d)[r:(%d,%d)]", 
				((DividerPosition*)currentItem->val)->cell1,
				((DividerPosition*)currentItem->val)->cell2,
				((DividerPosition*)currentItem->rep->val)->cell1,
				((DividerPosition*)currentItem->rep->val)->cell2
			);
		while (currentItem->next != NULL) {
			currentItem = currentItem->next;
			printf(
				"->(%d,%d)[r:(%d,%d)]", 
				((DividerPosition*)currentItem->val)->cell1,
				((DividerPosition*)currentItem->val)->cell2,
				((DividerPosition*)currentItem->rep->val)->cell1,
				((DividerPosition*)currentItem->rep->val)->cell2
			);
		}
		printf("-| (n=%d)\n", set->size);
	}
}

void Set_add(Set* set, void* val, size_t valSize) {
	Item* item = malloc(sizeof(Item));
	Item_init(item, val, valSize);
	set->tail->next = item;
	item->rep = set->tail->rep;
	set->tail = item;
	++set->size;
}

Item* Set_peek(Set* set, int index){
	int i;
	Item* currentItem = set->head;
	for (
		i = 0; 
		i <= index && currentItem->next != NULL;
		++i && (currentItem = currentItem->next)
	);

	return currentItem;
}
void Set_remove(Set* set, void* val, int (*cmpfn)(void*, void*)) {
	Item* currentItem = set->head;
	if (cmpfn(currentItem->val, val) == 0 && currentItem->next != NULL) {
		Item* oldHead = set->head;
		set->head = currentItem->next;
		free(oldHead);
		--set->size;

		currentItem = set->head;

		do {
			currentItem->rep = set->head;
		} while (currentItem->next != NULL && (currentItem = currentItem->next));
		return;
	}

	while (currentItem->next != NULL) {
		if (cmpfn(currentItem->next->val, val) == 0) {
			if (currentItem->next->next != NULL) {
				Item* oldItem = currentItem->next;
				currentItem->next = currentItem->next->next;
				free(oldItem);
			} else {
				free(currentItem->next);
				set->tail = currentItem;
				currentItem->next = NULL;
			}
			--set->size;
			return;
		}
		currentItem = currentItem->next;
	}
}

bool Set_contains(Set* set, void* val, int (*cmpfn)(void*, void*)) {
	Item* currentItem = set->head;
	if (cmpfn(currentItem->val, val) == 0) {
		return true;
	}
	while (currentItem->next != NULL) {
		currentItem = currentItem->next;
		if (cmpfn(currentItem->val, val) == 0) {
				return true;
		}
	}
	return false;
}
void Set_union(Set* set1, Set* set2) {
	// TODO: remove non unique values
	set1->tail->next = set2->head;
	set1->tail = set2->tail;
	set1->size += set2->size;
	
	Item* currentItem = set2->head;
	do {
		currentItem->rep = set1->head->rep;
	} while (currentItem->next != NULL && (currentItem = currentItem->next));
}
