#include "maze.h"
#include "djs.h"
#include "util.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

int dividerCmp(void* div1, void* div2) {
	// Invert the return value because the cmpfn is expected to return 0 if
	// equal, but for logical operations 0 == false and 1 == true

	return ! (
		((DividerPosition*)div1)->cell1 == ((DividerPosition*)div2)->cell1 
		&& ((DividerPosition*)div1)->cell2 == ((DividerPosition*)div2)->cell2
	);
}

int intCmp(void* int1, void* int2) {
	// Invert the return value because the cmpfn is expected to return 0 if
	// equal, but for logical operations 0 == false and 1 == true

	return ! (*(int*)int1 == *(int*)int2);
}

int printMaze(Set* removedDividers) {
	int i;
	int j;
	unsigned int cellNum;
	unsigned int rightCellNum = 0;
	unsigned int lowerCellNum = 0;
	DividerPosition divider;
	printf("\n[START]\n");
	putchar(' ');
	for (i = 0; i < MAZE_WIDTH-1; ++i) {
		if (i == 0) {
			printf("| |");
		} else {
			printf("___");
		}
	}
	printf("__");
	putchar('\n');
	for (i = 0; i < MAZE_WIDTH; ++i) {
		putchar('|');
		for (j = 0; j < MAZE_WIDTH; ++j) {
			cellNum = (MAZE_WIDTH * i) + j;
			if (cellNum < MAX_NUM_CELLS - MAZE_WIDTH) {
				lowerCellNum = cellNum + MAZE_WIDTH;
				divider.cell1 = cellNum;
				divider.cell2 = lowerCellNum;
				if (Set_contains(removedDividers, &divider, dividerCmp) == true) {
					printf("  ");
				} else {
					printf("__");
				}
			} else {
				printf("__");
			}

			if ((cellNum + 1) % MAZE_WIDTH != 0) {
				rightCellNum = cellNum + 1;
				divider.cell1 = cellNum;
				divider.cell2 = rightCellNum;
				if (Set_contains(removedDividers, &divider, dividerCmp) == true) {
					printf(" ");
				} else {
					printf("|");
				}
			} else {
				if (i < MAZE_WIDTH -1) {
					printf("|");
				} else {
					printf(" [END]");
				}
			}
			//printf("_%d|", cellNum);
		}
		putchar('\n');
	}
}


Set* genMaze(Set* cells, Set* dividers) {
	printf("Generating Maze...\n");
	Set* removedDividers = malloc(sizeof(Set));
	removedDividers->size = 0;

	size_t currNumCells = MAX_NUM_CELLS;

	int i;
	for (i = 0; i < MAX_NUM_CELLS; ++i) {
		int* nodeNum = malloc(sizeof(int));
		*nodeNum = i;	
		Set_createSet(cells+i, nodeNum, sizeof(int));
	}

	DividerPosition* rightDivider = NULL;
	DividerPosition* bottomDivider = NULL;
	Set* tempSet = malloc(sizeof(Set));

	for (i = 0; i < MAX_NUM_CELLS; ++i) {
		if (i == 0) {
			bottomDivider = malloc(sizeof(DividerPosition));
			bottomDivider->cell1 = i;
			bottomDivider->cell2 = i + MAZE_WIDTH;
			Set_createSet(dividers, bottomDivider, sizeof(DividerPosition));

			rightDivider = malloc(sizeof(DividerPosition));
			rightDivider->cell1 = i;
			rightDivider->cell2 = i + 1;
			Set_add(dividers, rightDivider, sizeof(DividerPosition));
			continue;
		}
		if (i < (MAX_NUM_CELLS - MAZE_WIDTH)) {
			
			bottomDivider = malloc(sizeof(DividerPosition));
			bottomDivider->cell1 = i;
			bottomDivider->cell2 = i + MAZE_WIDTH;
			Set_add(dividers, bottomDivider, sizeof(DividerPosition));
		}
		if ((i+1) % MAZE_WIDTH != 0) {
			rightDivider = malloc(sizeof(DividerPosition));
			rightDivider->cell1 = i;
			rightDivider->cell2 = i + 1;
			Set_add(dividers, rightDivider, sizeof(DividerPosition));
		}
	}
	
	long int randIndex;
	enum Direction direction;
	unsigned int cell1, cell2;
	int j;
	int set2Index;
	DividerPosition* selectedDivider;
	Set* set1;
	Set* set2;
	while (removedDividers->size < MAX_NUM_CELLS - 1) {
		set1 = NULL;
		set2 = NULL;
		randIndex = randInt() % dividers->size;
		selectedDivider = (DividerPosition*)Set_peek(dividers, randIndex)->val;
		cell1 = selectedDivider->cell1;
		cell2 = selectedDivider->cell2;
		for (j = 0; j < currNumCells; ++j) {
			if (Set_contains(cells+j, &cell1, &intCmp)) {
				set1 = cells+j;
			}
			if (Set_contains(cells+j, &cell2, &intCmp)) {
				set2 = cells+j;
				set2Index = j;
			}
		}

		if (set1 != NULL && set2 != NULL && set1->head->rep != set2->head->rep) {
			Set_remove(dividers, selectedDivider, &dividerCmp);
			if (removedDividers->size == 0) {
				Set_createSet(
					removedDividers,
					selectedDivider,
					sizeof(DividerPosition)
				);
			} else {
				Set_add(
					removedDividers,
					selectedDivider,
					sizeof(DividerPosition)
				);
			}
			Set_union(set1, set2);
			array_remove(cells, set2Index, currNumCells, sizeof(Set));
			--currNumCells;
		}

	}

	printf("Done!\n");
	printf("\n-- Removed Dividers--\n");
	Set_printMembers(removedDividers, true);
	printf("\n-- Cells --\n");
	for (i = 0; i < currNumCells; ++i) {
		Set_printMembers(cells+i, false);
	}

	return removedDividers;
}

int main() {
	Set* cells = calloc(MAX_NUM_CELLS, sizeof(Set));
	Set* dividers = malloc(sizeof(Set));
	Set* removedDividers = genMaze(cells, dividers);

	printMaze(removedDividers);
}
